import numpy as np


def to_vec(*, lat, lon):
    cos_lat = np.cos(lat)
    cos_lon = np.cos(lon)
    sin_lat = np.sin(lat)
    sin_lon = np.sin(lon)
    return np.stack([cos_lat * cos_lon, cos_lat * sin_lon, sin_lat], axis=-1)


def batched_dot(a, b):
    a = np.expand_dims(a, 1)
    b = np.expand_dims(b, 2)
    return np.matmul(a, b).squeeze(-1).squeeze(-1)


def cos_distance(*, lat1, lat2, dlon):
    cos_lat1 = np.cos(lat1)
    sin_lat1 = np.sin(lat1)

    cos_lat2 = np.cos(lat2)
    sin_lat2 = np.sin(lat2)

    cos_dlon = np.cos(dlon)

    return sin_lat1 * sin_lat2 + cos_lat1 * cos_lat2 * cos_dlon


def cos_distance_segment(lat, lon, *, lat1, lon1, lat2, lon2, eps=1e-5):
    x = to_vec(lat=lat, lon=lon)
    a, b = to_vec(
        lat=np.stack([lat1, lat2], axis=0), lon=np.stack([lon1, lon2], axis=0)
    )

    n = np.cross(a, b)
    cos_gamma = np.dot(a, b)
    n /= max(np.sqrt((1.0 - cos_gamma) * (1.0 + cos_gamma)), np.nextafter(0.0, 1.0))

    s = x @ n

    sel = np.abs(s) < 1.0 - np.pi**2 / 8.0 * eps**2  # rel. error < eps
    s = np.expand_dims(s[sel], -1)

    c = np.ones_like(x) * b
    c[sel] = (x[sel] - np.outer(s, n)) / np.sqrt((1.0 - s) * (1.0 + s))

    cos_alpha = c @ b
    cos_beta = c @ a

    closest_point = np.ones_like(x) * b

    sel = cos_gamma < np.min((cos_alpha, cos_beta), axis=0)  # gamma < max(alpha, beta)
    closest_point[sel] = c[sel]

    sel = ~sel & (cos_beta >= cos_alpha)  # beta <= alpha
    closest_point[sel] = a

    return batched_dot(x, closest_point)



import numpy as np

from geordpy import rdp_filter


def run_example():
    points = [
        (0, 0),
        (-10, 10),
        (10, 20),
        (-20, 30),
        (20, 40),
        (-10, 50),
        (10, 60),
        (0, 70),
    ]

    radius = 1000

    mask = rdp_filter(points, threshold=0, radius=radius)
    assert all(mask)

    mask = rdp_filter(points, threshold=350, radius=radius)
    assert all(mask == [True, False, False, False, False, False, False, True])

    threshold = 346
    mask = rdp_filter(points, threshold=threshold, radius=radius)
    assert all(mask == [True, False, False, True, True, False, False, True])

    print(np.array(points)[mask])


if __name__ == "__main___":
    run_example()  # pragma: no cover



import numpy as np

from geordpy import great_circle


def _filter(points, threshold):
    n_points = points.shape[0]
    if n_points <= 2:
        return np.full(n_points, True)

    mask = np.full(n_points, True)
    stack = [(0, n_points - 1)]

    while stack:
        first, last = stack.pop()
        if last - first < 2:
            continue

        dist = 1.0 - great_circle.cos_distance_segment(
            points[first + 1 : last, 0],
            points[first + 1 : last, 1],
            lat1=points[first, 0],
            lon1=points[first, 1],
            lat2=points[last, 0],
            lon2=points[last, 1],
        )
        i_max = np.argmax(dist) + 1  # dist[i] = dist(points[i+1], line seg.)
        dist_max = dist[i_max - 1]

        if dist_max < threshold:
            mask[first + 1 : last] = False
        else:
            stack.append((first, first + i_max))
            stack.append((first + i_max, last))

    return mask


def rdp_filter(points, threshold, radius=6_371_000):
    """
    Simplify a geodetic-coordinate polyline using the Ramer-Douglas-Peucker algorithm.

    This function applies the Ramer-Douglas-Peucker (RDP) algorithm to a list of geodetic-coordinate points,
    aiming to simplify the polyline while keeping the error below a specified threshold. The algorithm
    works by approximating the original polyline with a reduced number of points that lie close to the
    original curve.

    Args:
        points (list of tuple): A list of latitude and longitude pairs (given in degrees) representing the
                               geodetic-coordinate polyline points to be simplified.

        threshold (float): The maximum allowable error, specified as an arc length along great circle segments.
                          Points that deviate from the simplified curve by more than this threshold will be kept.

        radius (float, optional): The radius of the sphere used for calculations. Defaults to Earth's mean radius
                                 in meters (6371000). The threshold is interpreted based on this radius.

    Returns:
        list of bool: A binary mask indicating whether each point in the input list should be kept or discarded
                      based on the RDP simplification. The mask has the same length as the input 'points' list,
                      where a value of True indicates that the corresponding point should be kept, and False
                      indicates that the point can be discarded.

    Note:
        - The input 'points' list should have at least two points to define a valid polyline.
        - The 'threshold' value must be greater than zero.
        - The 'radius' value should be set appropriately to match the units of the threshold. For example, if the
          radius is set to 1, then the threshold corresponds to the arc length on the unit sphere.
    """
    if len(points) == 0:
        return np.empty(0, dtype=bool)

    points = np.deg2rad(np.array(points))
    threshold = 1.0 - np.cos(threshold / radius)  # negate to make it a distance

    return _filter(points, threshold)



#!/usr/bin/env python
# coding=utf-8

import pathlib
import sys

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()
sys.path.append(here)

import versioneer  # noqa: E402

with open(here / "README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="geordpy",
    description="A Python library for simplifying geodetic-coordinate polylines using the Ramer-Douglas-Peucker algorithm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avitase/geordpy",
    author="Nis Meinert",
    author_email="mail@nismeinert.name",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(exclude=["docs", "tests"]),
    python_requires=">=3.10",
    install_requires=[
        "numpy",
    ],
    extras_require={
        "dev": [
            "pre-commit",
            "pytest",
            "pytest-cov",
            "sphinx",
            "scipy",
        ],
    },
)
